# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Export directory to various formats."""

from __future__ import annotations

import logging

import sys

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources

import jinja2

from mots.directory import Directory
from mots.settings import settings

logger = logging.getLogger(__name__)

DEFAULT_LIST_TABLE_INDENT = 8


def escape_for_rst(value: str) -> str:
    """Escape rst special characters."""
    # First we must escape backslashes, then everything else.
    characters = ("\\", "*", "`")
    for character in characters:
        value = value.replace(character, f"\\{character}")
    return value


def escape_for_md(value: str) -> str:
    """Escape rst special characters."""
    # First we must escape backslashes, then everything else.
    characters = "\\`*_{}[]<>()#+-.!|"
    for character in characters:
        value = value.replace(character, f"\\{character}")
    return value


def format_link_for_rst(text: str, url: str) -> str:
    """Format a link."""
    return f"`{text} <{url}>`__"


def format_link_for_md(text: str, url: str) -> str:
    """Format a link."""
    return f"[{text}]({url})"


def format_paths_for_rst(
    value: list[str], indent: int, directory: Directory = None
) -> str:
    """Escape and format a path string (e.g. for includes or excludes)."""
    config = directory.config_handle.config
    try:
        searchfox_enabled = config["export"]["searchfox_enabled"]
    except KeyError:
        searchfox_enabled = False

    parsed_paths = []
    for path in value:
        path = path.replace("*", "\\*")
        if searchfox_enabled:
            path = format_link_for_rst(
                path,
                f"{settings.SEARCHFOX_BASE_URL}/{config['repo']}/search?q=&path={path}",
            )
        parsed_paths.append(path)
    return f"\n{' ' * indent}| " + f"\n{' ' * indent}| ".join(parsed_paths)


def format_paths_for_md(
    value: list[str], indent: int, directory: Directory = None
) -> str:
    """Escape and format a path string (e.g. for includes or excludes)."""
    config = directory.config_handle.config
    try:
        searchfox_enabled = config["export"]["searchfox_enabled"]
    except KeyError:
        searchfox_enabled = False

    parsed_paths = []
    for path in value:
        path = path.replace("*", "\\*")
        if searchfox_enabled:
            path = format_link_for_md(
                path,
                f"{settings.SEARCHFOX_BASE_URL}/{config['repo']}/search?q=&path={path}",
            )
        parsed_paths.append(path)
    return f"\n{' ' * indent}* " + f"\n{' ' * indent}* ".join(parsed_paths)


def format_people_for_rst(value: list[dict], indent: int) -> str:
    """Escape and format a list of people."""
    people_base_url = settings.PMO_SEARCH_URL
    parsed_people = []
    for person in value:
        url = f"{people_base_url}{person['nick']}"
        if "name" in person and person["name"]:
            parsed_person = format_link_for_rst(
                f"{person['name']} ({person['nick']})", url
            )
        else:
            parsed_person = format_link_for_rst(person["nick"], url)

        parsed_people.append(parsed_person)
    return f"\n{' ' * indent}| " + f"\n{' ' * indent}| ".join(parsed_people)


def format_people_for_md(value: list[dict], indent: int) -> str:
    """Escape and format a list of people."""
    people_base_url = settings.PMO_SEARCH_URL
    parsed_people = []
    for person in value:
        url = f"{people_base_url}{person['nick']}"
        if "name" in person and person["name"]:
            parsed_person = format_link_for_md(
                f"{person['name']} ({person['nick']})", url
            )
        else:
            parsed_person = format_link_for_md(person["nick"], url)

        parsed_people.append(parsed_person)
    return f"\n{' ' * indent}* " + f"\n{' ' * indent}* ".join(parsed_people)


def format_emeritus(value: list[dict | str]) -> str:
    """Return names of people if provided a mixed list."""
    parsed = []
    for person in value:
        if isinstance(person, str):
            parsed.append(person)
        elif isinstance(person, dict):
            if "name" in person and person["name"]:
                parsed.append(person["name"])
            elif "nick" in person and person["nick"]:
                parsed.append(person["nick"])
    return ", ".join(parsed)


def format_review_group_for_rst(group: str) -> str:
    """Format a review group."""
    base_url = settings.PHABRICATOR_BASE_URL
    return format_link_for_rst(f"#{group}", f"{base_url}/tag/{group}/")


def format_review_group_for_md(group: str) -> str:
    """Format a review group."""
    base_url = settings.PHABRICATOR_BASE_URL
    return format_link_for_md(f"#{group}", f"{base_url}/tag/{group}/")


class Exporter:
    """A helper class that exports to various formats."""

    @staticmethod
    def _get_env() -> jinja2.Environment:
        """Return a Jinja2 environment preloaded with a FileSystemLoader."""
        loader = jinja2.FileSystemLoader(
            searchpath=importlib_resources.files("mots") / "templates"
        )
        env = jinja2.Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
        env.filters["escape_for_rst"] = escape_for_rst
        env.filters["format_paths_for_rst"] = format_paths_for_rst
        env.filters["format_people_for_rst"] = format_people_for_rst
        env.filters["format_review_group_for_rst"] = format_review_group_for_rst
        env.filters["escape_for_md"] = escape_for_md
        env.filters["format_paths_for_md"] = format_paths_for_md
        env.filters["format_people_for_md"] = format_people_for_md
        env.filters["format_review_group_for_md"] = format_review_group_for_md
        env.filters["format_emeritus"] = format_emeritus
        return env

    def __init__(self, directory: Directory):
        self.directory = directory
        self.env = self._get_env()

    def _get_template(self, frmt: str) -> jinja2.Template:
        """Get a template based on the provided format."""
        template = self.env.get_template(f"directory.template.{frmt}")
        return template

    def _export_to_rst(self) -> str:
        """Render the template with this instance's directory context and return it."""
        template = self._get_template("rst")
        out = template.render(directory=self.directory)
        return f"{out.strip()}\n"

    def _export_to_md(self) -> str:
        """Render the template with this instance's directory context and return it."""
        template = self._get_template("md")
        out = template.render(directory=self.directory)
        return f"{out.strip()}\n"


def export_to_format(directory: Directory, frmt="rst"):
    """Export directory in a specified format."""
    supported_formats = ["rst", "md"]
    if frmt not in supported_formats:
        raise ValueError(f"{frmt} not one of {supported_formats}.")

    exporter = Exporter(directory)
    return getattr(exporter, f"_export_to_{frmt}")()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Configuration classes used to initialize and manage mots in a repo."""

from __future__ import annotations

from collections import defaultdict
import io
import hashlib
import logging

from datetime import datetime, timezone
from pathlib import Path
from ruamel.yaml import YAML

from mots.bmo import get_bmo_data
from mots.directory import Directory, People
from mots.export import export_to_format
from mots.module import Module
from mots.utils import generate_machine_readable_name
from mots.settings import settings

logger = logging.getLogger(__name__)

yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)

QUICK_START_BLURB = "\n".join(
    (
        "See https://mots.readthedocs.io/en/latest/#quick-start for quick start",
        "documentation and how to modify this file.",
    )
)

MPL2 = "\n".join(
    (
        "This Source Code Form is subject to the terms of the Mozilla Public",
        "License, v. 2.0. If a copy of the MPL was not distributed with this",
        "file, You can obtain one at https://mozilla.org/MPL/2.0/.",
    )
)


class ValidationError(TypeError):
    """Thrown when a particular module is not valid."""

    pass


class FileConfig:
    """Loader and writer for filesystem based configuration."""

    def __init__(self, path: Path = settings.DEFAULT_CONFIG_FILEPATH):
        """Initialize the configuration with provided config file path."""
        if not path.exists() and path.is_file():
            raise ValueError(f"{path} does not exist or is not a file.")
        self.path = path
        self.repo_path = path.parent
        self.config = None

    def init(self):
        """Initialize a repo with a config file, if it does not contain it."""
        if not self.path.is_file():
            # File does not exist, create it.
            now = datetime.now(timezone.utc).isoformat()
            self.config = {
                "repo": str(Path(self.path).resolve().parent.name),
                "created_at": now,
                "updated_at": None,
                "hashes": {"config": None, "export": None},
                "export": {},
                "people": [],
                "modules": [],
            }
            self.write()
            logger.info(f"mots configuration initialized in {self.path}.")
        else:
            logger.warning(f"mots configuration file detected in {self.path}.")

    def load(self):
        """Load configuration from file."""
        with self.path.open("r", encoding="utf-8") as f:
            self.config = yaml.load(f)

    def check_hashes(self) -> list[str]:
        """Check that the hashes in the config are up to date.

        Upon calling this function, the existing configuration is copied and stripped
        of volatile keys, then a hash is calculated and compared against the old hash.

        If there is a mismatch, return non-zero exit code. Otherwise return 0.
        """
        errors = []
        config = self.config.copy()

        if "export" in self.config and "path" in self.config["export"]:
            with (self.repo_path / config["export"]["path"]).open("rb") as f:
                export = f.read()
        else:
            export = None

        original_hashes, hashes = calculate_hashes(self.config, export)

        for hash_key in ("config", "export"):
            if original_hashes.get(hash_key) != hashes.get(hash_key):
                errors.append(f"Mismatch in {hash_key} hash detected.")
                errors.append(
                    f"{hashes[hash_key]} does not match {original_hashes[hash_key]}"
                )
                errors.append(f"{hash_key} file is out of date.")

        return errors

    def write(self, hashes: dict | None = None):
        """Write configuration to file, and update the timestamp and hashes."""
        logger.debug(f"Writing configuration to {self.path}")
        self.config["updated_at"] = datetime.now(timezone.utc).isoformat()
        self.config["hashes"] = hashes or {}
        with self.path.open("w", encoding="utf-8", newline="\n") as f:
            yaml.dump(self.config, f)


def reference_anchor_for_module(
    index: int,
    person: dict,
    key: str,
    file_config: FileConfig,
    directory: Directory,
    module: dict,
) -> None:
    """Associate person with a reference to a directory entry if possible.

    :param index: the position of the person in the referrer field
    :param person: a dictionary with information about the person to be added
    :param key: the key (e.g, "peers" or "owners_emeritus") that defines context
    :param file_config: config context being modified
    :param module: dictionary of the module in the config being modified


    This is needed so that ruamel.yaml can correctly reference anchors to people.
    """
    logger.debug(f"Setting reference to {person} as {key} in {module['machine_name']}")

    referrer = module["meta"][key] if key.endswith("_emeritus") else module[key]

    if person["bmo_id"] not in directory.people.by_bmo_id:
        # Person has to be added to directory before adding the reference.
        file_config.config["people"].append(person)
        directory.load_people()
        directory.people.refresh_by_bmo_id()
        referrer[index] = person
    else:
        # Associate the directory entry with the referrer.
        referrer[index] = file_config.config["people"][
            directory.people.by_bmo_id[person["bmo_id"]]
        ]


def calculate_hashes(config: dict, export: bytes) -> tuple[dict, dict]:
    """Calculate a hash of the yaml config file."""
    config = config.copy()

    # Exclude hashes and updated timestamp from hash generation
    original_hashes = config.pop("hashes", {})
    hashes = {}
    config.pop("updated_at", None)

    # Write actual config yaml dump to stream.
    with io.StringIO() as stream:
        yaml.dump(config, stream)
        content = stream.getvalue()

    config_hash = hashlib.sha1(content.encode("utf-8")).hexdigest()
    hashes["config"] = config_hash

    if "export" in config:
        hashes["export"] = hashlib.sha1(export).hexdigest()

    return original_hashes, hashes


def clean(file_config: FileConfig, write: bool = True, refresh: bool = True):
    """Clean and re-sort configuration file.

    Load configuration from disk, sort modules and submodules by `machine_name`. If
    there is no valid `machine_name`, generate one. Reformat yaml content. Calculate
    and store hashes if needed. Write changes to disk if needed.

    :param file_config: an instance of :class:`FileConfig`
    :param write: if set to `True`, writes changes to disk.
    """
    people_keys = ("owners", "peers")
    emeritus_keys = [f"{k}_emeritus" for k in people_keys]
    file_config.load()
    directory = Directory(file_config)
    directory.load()

    people = list(file_config.config["people"])

    bmo_data = get_bmo_data(people)
    updated_people = People(people, bmo_data)

    people_to_sync = set(person["bmo_id"] for person in people if "nick" not in person)

    if refresh:
        # Use the updated list that was synchronized with Bugzilla.
        logger.info("Refreshing all people entries from Bugzilla.")
        file_config.config["people"] = updated_people.serialized
    else:
        # Use the original people list, and update entries only where needed.
        logger.warning("Only synchronizing new people with Bugzilla.")
        updated_people_dict = {
            updated_person["bmo_id"]: updated_person
            for updated_person in updated_people.serialized
        }
        for person in people:
            if person["bmo_id"] in people_to_sync:
                logger.info(f"Updated {person['bmo_id']} with new data.")
                person.update(updated_people_dict[person["bmo_id"]])
        file_config.config["people"] = people

    for i, module in enumerate(file_config.config["modules"]):
        if "machine_name" not in module:
            module["machine_name"] = generate_machine_readable_name(module["name"])

        for emeritus_key in emeritus_keys:
            if not (
                "meta" in module
                and module["meta"]
                and emeritus_key in module["meta"]
                and module["meta"][emeritus_key]
            ):
                continue
            for i, person in enumerate(module["meta"][emeritus_key]):
                if not isinstance(person, dict):
                    continue
                reference_anchor_for_module(
                    i, person, emeritus_key, file_config, directory, module
                )

        for key in people_keys:
            if key not in module or not module[key]:
                continue
            for i, person in enumerate(module[key]):
                reference_anchor_for_module(
                    i, person, key, file_config, directory, module
                )

        # Do the same for submodules.
        if "submodules" in module and module["submodules"]:
            module["submodules"].sort(key=lambda x: x["name"])
            for submodule in module["submodules"]:

                for emeritus_key in emeritus_keys:
                    if not (
                        "meta" in submodule
                        and submodule["meta"]
                        and emeritus_key in submodule["meta"]
                        and submodule["meta"][emeritus_key]
                    ):
                        continue
                    for i, person in enumerate(submodule["meta"][emeritus_key]):
                        if not isinstance(person, dict):
                            continue
                        reference_anchor_for_module(
                            i,
                            person,
                            emeritus_key,
                            file_config,
                            directory,
                            submodule,
                        )

                for key in people_keys:
                    if key not in submodule or not submodule[key]:
                        continue
                    for i, person in enumerate(submodule[key]):
                        reference_anchor_for_module(
                            i, person, key, file_config, directory, submodule
                        )
                if "machine_name" not in submodule:
                    submodule["machine_name"] = generate_machine_readable_name(
                        submodule["name"]
                    )

    file_config.config["modules"].sort(key=lambda x: x["machine_name"])
    file_config.config.yaml_set_start_comment(f"{MPL2}\n\n{QUICK_START_BLURB}")
    if write:
        # Write all changes.
        file_config.write()

        # Reload changes from disk so we can reset all anchors.
        file_config.load()

        nicks = []
        file_config.config["people"].sort(
            key=lambda person: person.get("nick", "").lower()
        )
        for person in file_config.config["people"]:
            machine_readable_nick = generate_machine_readable_name(
                person.get("nick", ""), keep_case=True
            )
            if machine_readable_nick in nicks or not machine_readable_nick:
                continue
            nicks.append(machine_readable_nick)
            person.yaml_set_anchor(machine_readable_nick)

        if "export" in file_config.config and "format" in file_config.config["export"]:
            export = export_to_format(
                directory, file_config.config["export"]["format"]
            ).encode("utf-8")
        else:
            export = None
        hashes = calculate_hashes(file_config.config, export)[1]
        file_config.write(hashes)


def validate(config: dict, repo_path: str) -> list[str] | None:
    """Validate the current state of the config file.

    - Check if top-level dictionary contains required keys
    - Check if machine names are unique
    - Instantiate and run validation on each :class:`Module` instance

    :raises ValidationError: if any validation errors are detected
    """
    # Validate that config has all the required keys.
    required_keys = ("repo", "created_at", "updated_at", "modules")
    keys_diff = required_keys - config.keys()
    if len(keys_diff) != 0:
        raise ValidationError(f"{keys_diff} missing from configuration file.")

    modules = config["modules"]

    # Count machine name repetitions in a defaultdict.
    machine_names = defaultdict(int)
    for module in modules:
        machine_names[module["machine_name"]] += 1
        if "submodules" in module and module["submodules"]:
            for submodule in module["submodules"]:
                machine_names[submodule["machine_name"]] += 1

    machine_names = {name: count for name, count in machine_names.items() if count > 1}
    if machine_names:
        error_msg = (
            f"Duplicate machine name(s) found: {', '.join(machine_names.keys())}"
        )
        return [error_msg]

    errors = []
    for module in config["modules"]:
        module = Module(repo_path=repo_path, **module)
        validation_errors = module.validate()
        if validation_errors:
            errors += validation_errors

    if errors:
        return errors
    logger.info("All modules validated successfully.")


def add(
    new_module: dict, file_config: FileConfig, parent: str = None, write: bool = True
):
    """Add a new module to the configuration.

    :param module: a dictionary containing module parameters
    :param file_config: an instance of :class:`FileConfig`
    :param parent: the machine name of the parent module if applicable
    :param write: if set to `True`, writes changes to disk
    """
    file_config.load()
    modules = file_config.config["modules"]
    serialized = Module(**new_module, repo_path=file_config.repo_path).serialize()

    if parent:
        for module in modules:
            if module["machine_name"] == parent:
                if "submodules" not in module or not module["submodules"]:
                    module["submodules"] = []
                module["submodules"].append(serialized)
                break
    else:
        modules.append(serialized)

    if write:
        file_config.write()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Test configuration and fixtures."""

import pytest


@pytest.fixture
def repo(tmp_path, config):
    """Create a file structure in the file system and initialize a repo config.

    The below directory and file structure was created to provide different
    opportunities for categorizing the paths within each directory. This was done in
    the hope that it will make it easier to understand the different module definitions
    in other fixtures and tests.
    """
    from mots.config import FileConfig

    test_repo = tmp_path / "test_repo"
    test_repo.mkdir()

    [
        (test_repo / ".hg").mkdir(),
        (test_repo / "canines").mkdir(),
        (test_repo / "canines" / "chihuahuas").mkdir(),
        (test_repo / "canines" / "chihuahuas" / "apple_head").touch(),
        (test_repo / "canines" / "beagle").touch(),
        (test_repo / "canines" / "corgy").touch(),
        (test_repo / "canines" / "red_fox").touch(),
        (test_repo / "canines" / "hyena").touch(),
        (test_repo / "felines").mkdir(),
        (test_repo / "felines" / "persian").touch(),
        (test_repo / "felines" / "cheetah").touch(),
        (test_repo / "bovines").mkdir(),
        (test_repo / "bovines" / "cow").touch(),
        (test_repo / "bovines" / "sheep").touch(),
        (test_repo / "pigs").mkdir(),
        (test_repo / "pigs" / "wild_boar").touch(),
        (test_repo / "pigs" / "miniature_pig").touch(),
        (test_repo / "marsupials").mkdir(),
        (test_repo / "marsupials" / "kangaroo").touch(),
        (test_repo / "marsupials" / "koala").touch(),
        (test_repo / "birds").mkdir(),
        (test_repo / "birds" / "parrot").touch(),
        (test_repo / "birds" / "eagle").touch(),
        (test_repo / "mots.rst").touch(),
    ]

    file_config = FileConfig(test_repo / "mots.yml")
    file_config.config = config
    hashes = {
        "config": "f14a84e9e7a9f39ece7ac7232e2f55dda4da6e54",
        "export": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
    }
    file_config.write(hashes)

    return test_repo


@pytest.fixture
def config():
    people = [
        {"name": "jane", "nick": "jane", "bmo_id": 0},
        {"name": "jill", "nick": "jill", "bmo_id": 1},
        {"name": "otis", "nick": "otis", "bmo_id": 2},
        {"name": "angel", "nick": "angel", "bmo_id": 4},
    ]
    return {
        "repo": "test_repo",
        "created_at": "2021-09-10 12:53:22.383393",
        "updated_at": "2021-09-10 12:53:22.383393",
        "people": people,
        "export": {"format": "rst", "path": "mots.rst"},
        "modules": [
            {
                "machine_name": "domesticated_animals",
                "name": "Domesticated Animals",
                "exclude_submodule_paths": True,
                "exclude_module_paths": True,
                "includes": [
                    "canines/**/*",
                    "felines/**/*",
                    "bovines/**/*",
                    "birds/**/*",
                    "pigs/**/*",
                ],
                "excludes": ["canines/red_fox"],
                "owners": [people[0]],
                "peers": [people[1]],
                "meta": {"peers_emeritus": [people[2]]},
                "submodules": [
                    {
                        "machine_name": "predators",
                        "name": "Predators",
                        "includes": [
                            "canines/hyena",
                            "felines/tiger",
                            "felines/cheetah",
                            "birds/**/*",
                        ],
                        "excludes": [
                            "birds/parrot",
                        ],
                        "owners": [people[1]],
                        "meta": {"owners_emeritus": [people[2]]},
                    }
                ],
            },
            {
                "machine_name": "pets",
                "name": "Pets",
                "includes": [
                    "canines/**/*",
                    "felines/**/*",
                    "birds/**/*",
                ],
                "excludes": [
                    "canines/red_fox",
                    "canines/hyena",
                    "felines/cheetah",
                    "birds/eagle",
                ],
                "owners": [people[2]],
                "peers": [people[1]],
            },
        ],
    }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Integration tests for mots.module."""

from mots.config import FileConfig, add, validate
from mots.module import Module


def test_module__Module(repo):
    file_config = FileConfig(repo / "mots.yml")
    file_config.load()
    modules = file_config.config["modules"]

    assert len(modules) == 2
    m = Module(repo_path=repo, **modules[0])

    assert m.machine_name == "domesticated_animals"
    assert len(m.includes) == 5
    assert len(m.excludes) == 1
    assert len(m.submodules) == 1
    assert len(m.owners) == 1

    assert m.submodules[0].machine_name == "predators"
    assert len(m.submodules[0].includes) == 4
    # TODO: should excludes be inherited like this? Might need to revisit this.
    assert len(m.submodules[0].excludes) == 1
    assert len(m.submodules[0].owners) == 1


def test_module__Module__calculate_paths(repo):
    file_config = FileConfig(repo / "mots.yml")
    file_config.load()
    modules = file_config.config["modules"]
    m = Module(repo_path=repo, **modules[0])
    assert len(m.calculate_paths()) == 10
    assert len(m.submodules[0].calculate_paths()) == 3


def test_module__Module__validate(repo):
    m = Module(
        name="Some Module",
        machine_name="some_module",
        repo_path=str(repo),
        includes="*",
    )
    errors = m.validate()
    assert errors == []


def test_module__Module__validate__error_no_paths_in_submodule(repo):
    """Ensure an error is thrown when a submodule contains no valid paths."""
    m = dict(
        name="Some Module",
        machine_name="some_module",
        includes="*",
        submodules=[
            dict(
                name="Submodule",
                machine_name="submodule",
                excludes="*",
            ),
            dict(
                name="Submodule2",
                machine_name="submodule2",
                excludes="*",
            ),
        ],
    )
    m = Module(**m, repo_path=str(repo))
    errors = m.validate()
    assert errors == [
        "No valid paths were found in submodule.",
        "No valid paths were found in submodule2.",
    ]


def test_module__Module__submodule_with_no_peers_or_owners(repo):
    """Ensure a submodule with no peers or owners is parsed correctly."""
    m = dict(
        name="Some Module",
        machine_name="some_module",
        owners=[{"nick": "otis", "bmo_id": 2}],
        peers=[{"nick": "jill", "bmo_id": 1}],
        includes="*",
        submodules=[
            dict(
                name="Submodule",
                machine_name="submodule",
                excludes="some_path",
            ),
        ],
    )
    m = Module(**m, repo_path=str(repo))
    assert len(m.submodules[0].peers) == 0
    assert len(m.submodules[0].owners) == 0


def test_module__Module__validate__invalid_machine_name(repo):
    """Ensure an error is thrown when a submodule has an invalid machine name."""
    m = dict(
        machine_name="some module",
        includes="*",
    )
    m = Module(**m, repo_path=str(repo))
    errors = m.validate()
    assert errors == ["Machine name some module contains white space."]

    m = dict(
        machine_name="",
        includes="*",
    )
    m = Module(**m, repo_path=str(repo))
    errors = m.validate()
    assert errors == ["Module has a blank machine_name."]


def test_module__Module__serialize(repo):
    assert True


def test_module__validate(repo, config):
    result = validate(config, str(repo))
    assert result is None


def test_module__validate__error_duplicate_machine_names(repo):
    """Ensure an error is thrown when there are duplicate machine names in modules."""
    config = {
        "repo": "test_repo",
        "created_at": "2021-09-10 12:53:22.383393",
        "updated_at": "2021-09-10 12:53:22.383393",
        "modules": [
            {
                "name": "m1",
                "machine_name": "m1",
                "includes": ["*"],
                "submodules": [
                    {
                        "name": "m1",
                        "machine_name": "m1",
                    },
                    {
                        "name": "m3",
                        "machine_name": "m3",
                    },
                ],
            },
            {
                "name": "m2",
                "machine_name": "m2",
                "includes": ["*"],
                "submodules": [
                    {
                        "name": "m3",
                        "machine_name": "m3",
                    },
                ],
            },
        ],
    }

    messages = validate(config, repo)
    assert messages == ["Duplicate machine name(s) found: m1, m3"]


def test_module__validate__error_no_paths(repo):
    """Ensure an error is thrown when there are duplicate machine names in modules."""
    config = {
        "repo": "test_repo",
        "created_at": "2021-09-10 12:53:22.383393",
        "updated_at": "2021-09-10 12:53:22.383393",
        "modules": [
            {
                "name": "m1",
                "machine_name": "m1",
                "includes": ["does_not_exist"],
            },
        ],
    }

    messages = validate(config, repo)
    assert messages == ["No valid paths were found in m1."]


def test_module__add(repo):
    file_config = FileConfig(repo / "mots.yml")
    file_config.load()

    module = {
        "machine_name": "reptiles",
        "includes": ["reptiles/**/*"],
        "owners": [{"nick": "otis", "bmo_id": 2, "info": "testing"}],
    }

    add(module, file_config, parent="pets", write=True)
    file_config.load()
    # NOTE should we be storing machine names as keys in a dict, instead of a list?
    # even an ordered dict? This might make some things a little easier.
    assert (
        file_config.config["modules"][1]["submodules"][0]["machine_name"] == "reptiles"
    )
    assert file_config.config["modules"][1]["submodules"][0]["includes"] == [
        "reptiles/**/*"
    ]
    assert file_config.config["modules"][1]["submodules"][0]["owners"] == [
        {"nick": "otis", "bmo_id": 2, "info": "testing"}
    ]

    # TODO: validate that includes, excludes, and owners are lists...



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
This module provides an API to integrate mots within mach.

Be careful to not break backward compatibility here, or you will
break mach!
"""

from __future__ import absolute_import

from argparse import Namespace

from mots.cli import create_parser, main
from mots.utils import check_for_updates


parser = create_parser
new_release_on_pypi = check_for_updates


def run(options):
    """Run mots given a dict of options."""
    main(
        args=Namespace(**options),
        skip_update_check=True,
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for directory module."""

from mots.directory import Directory, Person, QueryResult
from mots.module import Module
from mots.config import FileConfig


def test_directory__Directory(repo):
    file_config = FileConfig(repo / "mots.yml")
    directory = Directory(file_config)
    directory.load(full_paths=True)

    rp = directory.repo_path
    di = directory.index

    assert len(di) == 24

    assert [m.machine_name for m in di[rp / "birds"]] == []
    assert [m.machine_name for m in di[rp / "birds/eagle"]] == ["predators"]
    assert [m.machine_name for m in di[rp / "birds/parrot"]] == ["pets"]
    assert [m.machine_name for m in di[rp / "bovines"]] == []
    assert [m.machine_name for m in di[rp / "bovines/cow"]] == ["domesticated_animals"]
    assert [m.machine_name for m in di[rp / "bovines/sheep"]] == [
        "domesticated_animals"
    ]
    assert [m.machine_name for m in di[rp / "canines"]] == []
    assert [m.machine_name for m in di[rp / "canines/beagle"]] == ["pets"]
    assert [m.machine_name for m in di[rp / "canines/chihuahuas"]] == ["pets"]
    assert [m.machine_name for m in di[rp / "canines/chihuahuas/apple_head"]] == [
        "pets"
    ]
    assert [m.machine_name for m in di[rp / "canines/corgy"]] == ["pets"]
    assert [m.machine_name for m in di[rp / "canines/hyena"]] == ["predators"]
    assert [m.machine_name for m in di[rp / "canines/red_fox"]] == []
    assert [m.machine_name for m in di[rp / "felines"]] == []
    assert [m.machine_name for m in di[rp / "felines/cheetah"]] == ["predators"]
    assert [m.machine_name for m in di[rp / "felines/persian"]] == ["pets"]
    assert [m.machine_name for m in di[rp / "marsupials"]] == []
    assert [m.machine_name for m in di[rp / "marsupials/kangaroo"]] == []
    assert [m.machine_name for m in di[rp / "marsupials/koala"]] == []
    assert [m.machine_name for m in di[rp / "mots.yml"]] == []
    assert [m.machine_name for m in di[rp / "pigs"]] == []
    assert [m.machine_name for m in di[rp / "pigs/miniature_pig"]] == [
        "domesticated_animals"
    ]
    assert [m.machine_name for m in di[rp / "pigs/wild_boar"]] == [
        "domesticated_animals"
    ]


def test_directory__Directory_new_path(repo):
    file_config = FileConfig(repo / "mots.yml")
    directory = Directory(file_config)
    directory.load()
    new_file_path = directory.repo_path / "canines/chihuahuas/deer_head"
    new_file_path.touch()
    assert new_file_path not in directory.index
    directory.load()
    assert new_file_path in directory.index
    assert len(directory.index[new_file_path]) == 1
    assert directory.index[new_file_path][0].machine_name == "pets"


def test_directory__Directory_deleted_path(repo):
    file_config = FileConfig(repo / "mots.yml")
    directory = Directory(file_config)
    directory.load()
    existing_path = directory.repo_path / "canines/chihuahuas/apple_head"
    assert len(directory.index[existing_path]) == 1
    assert directory.index[existing_path][0].machine_name == "pets"
    assert existing_path in directory.index
    existing_path.unlink()
    directory.load()
    assert existing_path not in directory.index


def test_directory__Directory__query(repo):
    file_config = FileConfig(repo / "mots.yml")
    directory = Directory(file_config)
    directory.load()

    paths_to_check = [
        "canines/chihuahuas/apple_head",
        "birds/parrot",
        "felines/persian",
        "felines/maine_coon",
    ]

    result = directory.query(*paths_to_check)
    assert result.path_map == {
        "canines/chihuahuas/apple_head": [directory.modules_by_machine_name["pets"]],
        "birds/parrot": [directory.modules_by_machine_name["pets"]],
        "felines/persian": [directory.modules_by_machine_name["pets"]],
    }
    assert set(result.paths) == {
        "canines/chihuahuas/apple_head",
        "birds/parrot",
        "felines/persian",
    }
    assert set(result.owners) == {Person(bmo_id=2, name="otis", nick="otis")}
    assert set(result.peers) == {Person(bmo_id=1, name="jill", nick="jill")}
    assert result.rejected_paths == ["felines/maine_coon"]


def test_directory__Directory__query_merging(repo):
    file_config = FileConfig(repo / "mots.yml")
    directory = Directory(file_config)
    directory.load()

    paths_to_check_1 = [
        "canines/chihuahuas/apple_head",
        "birds/parrot",
        "felines/maine_coon",
    ]
    paths_to_check_2 = [
        "felines/persian",
        "felines/maine_coon",
    ]

    result_1 = directory.query(*paths_to_check_1)
    result_2 = directory.query(*paths_to_check_2)

    result = result_1 + result_2

    assert result.path_map == {
        "canines/chihuahuas/apple_head": [directory.modules_by_machine_name["pets"]],
        "birds/parrot": [directory.modules_by_machine_name["pets"]],
        "felines/persian": [directory.modules_by_machine_name["pets"]],
    }
    assert set(result.paths) == {
        "canines/chihuahuas/apple_head",
        "birds/parrot",
        "felines/persian",
    }
    assert set(result.owners) == {Person(bmo_id=2, name="otis", nick="otis")}
    assert set(result.peers) == {Person(bmo_id=1, name="jill", nick="jill")}
    assert result.rejected_paths == ["felines/maine_coon"]


def test_directory__Directory__query_add_to_empty_QueryResult(repo):
    file_config = FileConfig(repo / "mots.yml")
    directory = Directory(file_config)
    directory.load()

    paths_to_check_1 = [
        "canines/chihuahuas/apple_head",
        "birds/parrot",
        "felines/maine_coon",
    ]

    result_1 = directory.query(*paths_to_check_1)

    result = result_1 + QueryResult()

    assert result.path_map == {
        "canines/chihuahuas/apple_head": [directory.modules_by_machine_name["pets"]],
        "birds/parrot": [directory.modules_by_machine_name["pets"]],
    }
    assert set(result.paths) == {
        "canines/chihuahuas/apple_head",
        "birds/parrot",
    }
    assert set(result.owners) == {Person(bmo_id=2, name="otis", nick="otis")}
    assert set(result.peers) == {Person(bmo_id=1, name="jill", nick="jill")}
    assert result.rejected_paths == ["felines/maine_coon"]


def test_directory__QueryResult_empty():
    empty_result = QueryResult()
    assert not empty_result


def test_directory__QueryResult_nonempty():
    test_module = Module(machine_name="test_module", repo_path="/repos/test")
    empty_result = QueryResult({"test_path": [test_module]}, ["rejected_path"])
    assert empty_result


def test_directory__QueryResult_empty_addition():
    empty_result = QueryResult()
    other_empty_result = QueryResult()
    assert not (empty_result + other_empty_result)


def test_directory__peers_and_owners(repo):
    file_config = FileConfig(repo / "mots.yml")
    directory = Directory(file_config)
    assert directory.peers_and_owners == [0, 1, 2]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Test config module."""

from operator import itemgetter
from unittest import mock

import pytest

from mots.config import (
    FileConfig,
    calculate_hashes,
    clean,
    reference_anchor_for_module,
)
from mots.directory import Directory


@pytest.fixture
def test_bmo_user_data():
    return {
        0: {"real_name": "janeway", "nick": "captain", "bmo_id": 0},
        1: {"real_name": "tuvok", "nick": "2vk", "bmo_id": 1},
        2: {"real_name": "neelix", "nick": "cooks4u", "bmo_id": 2},
        3: {"real_name": "seven", "nick": "nanotubes", "bmo_id": 3},
        4: {"real_name": "tom paris", "nick": "paris", "bmo_id": 4},
    }


def test_calculate_hashes(config):
    export = b""
    hashes = calculate_hashes(config, export)[1]

    assert (
        hashes["config"] == "f14a84e9e7a9f39ece7ac7232e2f55dda4da6e54"
    ), "Was `conftest.config` changed?"
    assert hashes["export"] == "da39a3ee5e6b4b0d3255bfef95601890afd80709"


def test_FileConfig__check_hashes(repo):
    file_config = FileConfig(repo / "mots.yml")
    file_config.load()
    errors = file_config.check_hashes()
    assert errors == []

    file_config.config["hashes"]["config"] = "asdf"
    file_config.config["hashes"]["export"] = "ghjk"
    errors = file_config.check_hashes()
    assert errors == [
        "Mismatch in config hash detected.",
        "f14a84e9e7a9f39ece7ac7232e2f55dda4da6e54 does not match asdf",
        "config file is out of date.",
        "Mismatch in export hash detected.",
        "da39a3ee5e6b4b0d3255bfef95601890afd80709 does not match ghjk",
        "export file is out of date.",
    ]


def test_reference_anchor_for_module(repo):
    """Test that a reference to an existing person is correctly updated.

    When a reference to a person references the same person but a different dictionary,
    reference_anchor_for_module should update that reference to reflect the existing
    person.
    """
    file_config = FileConfig(repo / "mots.yml")
    file_config.load()
    directory = Directory(file_config)
    directory.load(full_paths=True)

    module = file_config.config["modules"][0]["submodules"][0]
    owner_emeritus_field = module["meta"]["owners_emeritus"]
    # Check that existing owner emeritus references the person in the roster.
    assert len(owner_emeritus_field) == 1
    assert id(owner_emeritus_field[0]) == id(file_config.config["people"][2])

    # Change owner emeritus to be the first person in the roster.
    new_person = {"name": "jane", "nick": "jane", "bmo_id": 0}
    owner_emeritus_field[0] = new_person

    assert new_person == file_config.config["people"][0]
    assert id(owner_emeritus_field[0]) != id(file_config.config["people"][0])

    key = "owners_emeritus"
    index = 0
    reference_anchor_for_module(
        index, file_config.config["people"][0], key, file_config, directory, module
    )

    assert id(owner_emeritus_field[0]) == id(file_config.config["people"][0])


@mock.patch("mots.config.get_bmo_data")
def test_clean_removed_user_no_refresh(get_bmo_data, repo, config, test_bmo_user_data):
    """Test that removing a user without a refresh has no effect."""
    get_bmo_data.return_value = test_bmo_user_data

    file_config = FileConfig(repo / "mots.yml")
    file_config.load()

    assert file_config.config["people"] == config["people"]

    file_config.config["people"].pop()
    file_config.write()

    clean(file_config, refresh=False)

    cleaned_and_sorted = sorted(file_config.config["people"], key=itemgetter("bmo_id"))
    assert len(cleaned_and_sorted) == 3
    for i, person in enumerate(cleaned_and_sorted):
        assert person == file_config.config["people"][i]


@mock.patch("mots.config.get_bmo_data")
def test_clean_removed_user_with_refresh(
    get_bmo_data, repo, config, test_bmo_user_data
):
    """Test that removing a user with a refresh updates from BMO."""
    get_bmo_data.return_value = test_bmo_user_data

    file_config = FileConfig(repo / "mots.yml")
    file_config.load()

    assert file_config.config["people"] == config["people"]

    file_config.config["people"].pop()
    file_config.write()

    clean(file_config, refresh=True)

    cleaned_and_sorted = sorted(file_config.config["people"], key=itemgetter("bmo_id"))
    assert len(cleaned_and_sorted) == 3
    for i, person in enumerate(cleaned_and_sorted):
        assert person["name"] == test_bmo_user_data[i]["real_name"]
        assert person["nick"] == test_bmo_user_data[i]["nick"]


@mock.patch("mots.config.get_bmo_data")
def test_clean_changed_names(get_bmo_data, repo, config, test_bmo_user_data):
    """Test that updated BMO data will be reflected when cleaning."""
    get_bmo_data.return_value = test_bmo_user_data

    file_config = FileConfig(repo / "mots.yml")
    file_config.load()

    assert file_config.config["people"] == config["people"]

    clean(file_config)

    for person in file_config.config["people"]:
        assert person["name"] == test_bmo_user_data[person["bmo_id"]]["real_name"]
        assert person["nick"] == test_bmo_user_data[person["bmo_id"]]["nick"]


@mock.patch("mots.config.get_bmo_data")
def test_clean_changed_names_no_refresh(get_bmo_data, repo, config, test_bmo_user_data):
    """Test that updated BMO data has no effect when cleaning without refresh."""
    get_bmo_data.return_value = test_bmo_user_data

    file_config = FileConfig(repo / "mots.yml")
    file_config.load()

    assert file_config.config["people"] == config["people"]

    clean(file_config, refresh=False)

    assert sorted(file_config.config["people"], key=itemgetter("bmo_id")) == sorted(
        config["people"], key=itemgetter("bmo_id")
    )


@mock.patch("mots.config.get_bmo_data")
def test_clean_added_user_no_refresh(get_bmo_data, repo, config, test_bmo_user_data):
    """Test that updated BMO data has no effect when cleaning without refresh."""
    get_bmo_data.return_value = test_bmo_user_data

    file_config = FileConfig(repo / "mots.yml")
    file_config.load()

    assert file_config.config["people"] == config["people"]

    file_config.config["people"].insert(1, {"bmo_id": 3})
    file_config.write()

    assert len(file_config.config["people"]) == 5

    # Check that the new entry is unaffected after write.
    for person in file_config.config["people"]:
        if person["bmo_id"] == 3:
            # Ensure only the bmo_id key is there.
            assert len(person) == 1, "Only the bmo_id key should be present."

    # Compare the old and new people lists without the new entry, they should match.
    new = sorted(
        [
            dict(person)
            for person in file_config.config["people"]
            if person["bmo_id"] != 3
        ],
        key=itemgetter("bmo_id"),
    )
    original = sorted([person for person in config["people"]], key=itemgetter("bmo_id"))
    assert new == original

    # Run clean without refresh.
    clean(file_config, refresh=False)

    # It is expected that only the new entry is updated, and the rest are intact.
    explain_assert_existing = "Existing records should not be updated after clean."
    cleaned = sorted(file_config.config["people"], key=itemgetter("bmo_id"))
    assert len(cleaned) == 5
    assert cleaned[0] == original[0], explain_assert_existing
    assert cleaned[1] == original[1], explain_assert_existing
    assert cleaned[2] == original[2], explain_assert_existing
    assert cleaned[3] == {
        "bmo_id": 3,
        "name": "seven",
        "nick": "nanotubes",
    }, "New entry should have been updated after clean."


@mock.patch("mots.config.get_bmo_data")
def test_clean_added_user_with_refresh(get_bmo_data, repo, config, test_bmo_user_data):
    """Test that updated BMO data is reflected when cleaning with a refresh."""
    get_bmo_data.return_value = test_bmo_user_data

    file_config = FileConfig(repo / "mots.yml")
    file_config.load()

    assert file_config.config["people"] == config["people"]

    file_config.config["people"].append({"bmo_id": 3})
    file_config.write()

    assert len(file_config.config["people"]) == 5

    # Check that the new entry is unaffected after write.
    for person in file_config.config["people"]:
        if person["bmo_id"] == 3:
            # Ensure only the bmo_id key is there.
            assert len(person) == 1

    # Compare the old and new people lists without the new entry, they should match.
    new = sorted(
        [
            dict(person)
            for person in file_config.config["people"]
            if person["bmo_id"] != 3
        ],
        key=itemgetter("bmo_id"),
    )
    original = sorted(config["people"], key=itemgetter("bmo_id"))
    assert new == original

    # Run clean with refresh.
    clean(file_config, refresh=True)

    # It is expected that all people will have been updated via BMO.
    explain_assert_all = "Record should have been updated after clean."
    cleaned = sorted(file_config.config["people"], key=itemgetter("bmo_id"))
    assert len(cleaned) == 5
    assert cleaned[0] == {
        "bmo_id": 0,
        "name": "janeway",
        "nick": "captain",
    }, explain_assert_all
    assert cleaned[1] == {
        "bmo_id": 1,
        "name": "tuvok",
        "nick": "2vk",
    }, explain_assert_all
    assert cleaned[2] == {
        "bmo_id": 2,
        "name": "neelix",
        "nick": "cooks4u",
    }, explain_assert_all
    assert cleaned[3] == {
        "bmo_id": 3,
        "name": "seven",
        "nick": "nanotubes",
    }, explain_assert_all
    assert cleaned[4] == {
        "bmo_id": 4,
        "name": "tom paris",
        "nick": "paris",
    }, explain_assert_all



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Python code style tests."""

import subprocess


def test_check_black():
    """Check that black runs cleanly and that no files require reformatting."""
    cmd = ("black", "--diff", "src/mots")
    output = subprocess.check_output(cmd)
    assert not output


def test_check_flake8():
    """Check that flake8 runs cleanly and does not produce any errors."""
    cmd = ("flake8", "--exit-zero", "tests", "src")
    output = subprocess.check_output(cmd)
    if output:
        raise AssertionError(output.decode("utf-8"))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Directory classes for mots."""
from __future__ import annotations

import copy
from collections import defaultdict
from dataclasses import asdict
from dataclasses import dataclass
from itertools import chain
import logging
from mots.module import Module
from mots.utils import parse_real_name

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mots.config import FileConfig


logger = logging.getLogger(__name__)


class Directory:
    """Path indexer.

    When this class is initialized with a :class:`FileConfig <mots.config.FileConfig>`
    instance, it parses the configuration and loads all modules into the instance. In
    order to use the rest of the methods provided by this class, the
    :func:`load <mots.directory.Directory.load>` method must be called first.
    """

    def __init__(self, config: "FileConfig"):
        self.config_handle = config
        self.reset_config()

        self.index = None
        self.people = None

    def reset_config(self):
        """Load config and refresh modules, etc..."""
        self.config_handle.load()

        self.repo_path = self.config_handle.repo_path
        self.modules = [
            Module(repo_path=self.repo_path, **m)
            for m in self.config_handle.config["modules"]
        ]

        self.modules_by_machine_name = {}
        for module in self.modules:
            self.modules_by_machine_name[module.machine_name] = module

        self.description = self.config_handle.config.get("description", "")

    def load(self, full_paths: bool = False):
        """Populate file path and people indexes.

        :param full_paths: when true, loads all repo paths in filesystem into index.

        This method should be called any time there are changes in the filesystem. For
        example, if the directory index is loaded before a patch is applied, then any
        new files that are added, removed, or moved, may not resolve correctly.
        """
        self.index = defaultdict(list)

        if full_paths:
            # Make sure all existing paths have an entry in the index.
            logger.debug("Loading paths into index...")
            for path in self.repo_path.glob("**/*"):
                if path.relative_to(self.repo_path).parts[0] in (".hg", ".git"):
                    continue
                self.index[path] = list()
            logger.debug(f"{len(self.index)} paths loaded.")

        for module in self.modules:
            if module.submodules:
                # Start with more specific submodule definitions if present.
                for submodule in module.submodules:
                    logger.debug(f"Updating index for {submodule.machine_name}...")
                    for path in submodule.calculate_paths():
                        self.index[path].append(submodule)
            logger.debug(f"Updating index for {module.machine_name}...")
            # Add broader module definitions.
            for path in module.calculate_paths():
                self.index[path].append(module)

        # Filter out modules that specifically exclude paths defined in other modules.
        for path in self.index.keys():
            modules = self.index[path]
            if len(modules) > 1:
                self.index[path] = [m for m in modules if not m.exclude_module_paths]
        self.index = dict(self.index)

        self.load_people()

    def load_people(self):
        """Load people directory from config handle."""
        people = list(self.config_handle.config["people"])
        self.people = People(people, {})

    def query(self, *paths: str) -> "QueryResult":
        """Query given paths and return a list of corresponding modules.

        :param paths: a string representing a path within the repo
        :rtype: mots.directory.QueryResult
        """
        result = {}
        rejected = []
        for path in paths:
            if not (self.repo_path / path).exists():
                logger.warning(f"Path {path} does not exist, skipping.")
                rejected.append(path)
                continue
            result[path] = self.index.get(self.repo_path / path, list())
        logger.debug(f"Query {paths} resolved to {result}.")

        return QueryResult(result, rejected)

    @property
    def peers_and_owners(self):
        """Return a sorted list of all peers and owners."""
        all_modules = self.modules
        all_submodules = list(chain(*[module.submodules for module in all_modules]))
        modules_and_submodules = all_modules + all_submodules

        peers_and_owners = set()
        for module in modules_and_submodules:
            peers_and_owners.update(
                [person["bmo_id"] for person in module.peers + module.owners]
            )
        return sorted(list(peers_and_owners))


class QueryResult:
    """Helper class to simplify query result interpretation."""

    data_keys = {
        "paths",
        "rejected_paths",
        "modules",
        "owners",
        "peers",
    }

    def __init__(
        self,
        result: dict[str, list[Module]] | None = None,
        rejected: list[str] | None = None,
    ):
        result = result or {}
        rejected = rejected or []

        data = {k: set() for k in self.data_keys}
        self.path_map = result

        for path in self.path_map:
            data["paths"].add(path)
            data["modules"].update(self.path_map[path])

        for module in data["modules"]:
            data["owners"].update([Person(**o) for o in module.owners])
            data["peers"].update([Person(**p) for p in module.peers])

        data["rejected_paths"].update(rejected)

        for key in data:
            setattr(self, key, list(data[key]))

    def __bool__(self):
        """Return whether there is any data in this object."""
        return bool(self.path_map)

    def __add__(self, query_result: "QueryResult") -> "QueryResult":
        """Merge the data from both QueryResult objects."""
        path_map = self.path_map.copy()
        path_map.update(query_result.path_map)
        rejected_paths = set.union(
            set(self.rejected_paths),
            set(query_result.rejected_paths),
        )
        return QueryResult(path_map, rejected_paths)

    def __radd__(self, query_result: "QueryResult") -> "QueryResult":
        """Call self.__add__ since the order of addition does not matter."""
        return self.__add__(query_result)


@dataclass
class Person:
    """A class representing a person."""

    bmo_id: int
    name: str
    nick: str

    def __hash__(self):
        """Return a unique identifier for this person."""
        return self.bmo_id


class People:
    """A people directory searchable by name, email, or BMO ID."""

    def __init__(self, people: list[dict], bmo_data: dict):
        logger.debug(f"Initializing people directory with {len(people)} people...")

        self.people = []
        self.by_bmo_id = {}

        people = copy.deepcopy(people)

        for i, person in enumerate(people):
            logger.debug(f"Adding person {person} to roster...")

            bmo_id = person["bmo_id"] = int(person["bmo_id"])
            if bmo_id in bmo_data and bmo_data[bmo_id]:
                # Update person's data base on BMO data.
                bmo_datum = bmo_data[person["bmo_id"]]
                person["nick"] = bmo_datum.get("nick", "")

                parsed_real_name = parse_real_name(bmo_datum["real_name"])
                person["name"] = parsed_real_name["name"]
            else:
                person["name"] = person.get("name", "")
                person["nick"] = person.get("nick", "")

            self.people.append(Person(**person))
            self.by_bmo_id[person["bmo_id"]] = i
            logger.debug(f"Person {person} added to position {i}.")
        self.serialized = [asdict(person) for person in self.people]

    def refresh_by_bmo_id(self):
        """Refresh index positions of people by their bugzilla ID."""
        self.by_bmo_id = {}
        for i, person in enumerate(self.people):
            self.by_bmo_id[person.bmo_id] = i



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Integration tests for settings."""

import pytest
from unittest.mock import patch
from mots.settings import Settings


@pytest.fixture
def mock_getenv():
    _vars = {
        "MOTS_KEY_2": "ENV_VALUE_2",
    }
    return lambda key, default=None: _vars.get(key, default)


@pytest.fixture
def mock_settings(monkeypatch):
    mock_defaults = {
        "KEY_1": "SET_VALUE_1",
        "KEY_2": "SET_VALUE_2",
        "KEY_3": "SET_VALUE_3",
    }
    monkeypatch.setattr(Settings, "DEFAULTS", mock_defaults)
    settings = Settings()
    return settings


def test__get_from_settings(mock_settings):
    assert mock_settings.KEY_1 == "SET_VALUE_1"
    assert mock_settings.KEY_2 == "SET_VALUE_2"
    assert mock_settings.KEY_3 == "SET_VALUE_3"


@patch("mots.settings.os.getenv")
def test__get_from_env(getenv, mock_settings, mock_getenv):
    getenv.side_effect = mock_getenv
    mock_settings.load()
    mock_settings.set_attributes()
    assert mock_settings.KEY_1 == "SET_VALUE_1"
    assert mock_settings.KEY_2 == "ENV_VALUE_2"
    assert mock_settings.KEY_3 == "SET_VALUE_3"


@patch("mots.settings.os.getenv")
def test__get_from_overrides(getenv, mock_settings, mock_getenv):
    getenv.side_effect = mock_getenv
    mock_settings.overrides = {
        "KEY_2": "OVERRIDE_VALUE_2",
        "KEY_3": "OVERRIDE_VALUE_3",
    }
    mock_settings.load()
    mock_settings.set_attributes()
    assert mock_settings.KEY_1 == "SET_VALUE_1"
    assert mock_settings.KEY_2 == "OVERRIDE_VALUE_2"
    assert mock_settings.KEY_3 == "OVERRIDE_VALUE_3"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Module that provides helpers to interact with the Bugzilla API."""

from __future__ import annotations

import requests
import logging

from mots.settings import settings


logger = logging.getLogger(__name__)


class MissingBugzillaAPIKey(Exception):
    """Raised when a Bugzilla API key was not detected in settings."""

    pass


def get_bmo_data(people: list) -> dict:
    """Fetch an updated dictionary from Bugzilla with user data.

    Dictionary keys are set to user IDs, and values are set to various data.
    """
    bmo_client = BMOClient()
    bmo_data = bmo_client.get_users_by_ids([p["bmo_id"] for p in people])
    return bmo_data


class BMOClient:
    """A thin wrapper as a Bugzilla API client."""

    def __init__(self, token: str | None = None, base_url: str = settings.BUGZILLA_URL):
        if not token:
            token = settings.BUGZILLA_API_KEY
            if not token:
                raise MissingBugzillaAPIKey()
        self._headers = {"X-BUGZILLA-API-KEY": token, "User-Agent": settings.USER_AGENT}
        self._base_url = base_url

    def _get(self, path: str, params=None):
        logger.debug(f"GET {self._base_url}/{path}")
        params = params or {}
        response = requests.get(
            f"{self._base_url}/{path}", headers=self._headers, params=params
        )
        return response

    def get_users_by_ids(self, ids: list[str]):
        """Get user data by BMO IDs."""
        fields = ["real_name", "nick", "name", "id", "email"]
        response = self._get("user", {"ids": ids, "include_fields": ",".join(fields)})
        if response.status_code != 200:
            logger.error(f"Error searching {ids}: {response.status_code}")
            return

        result = response.json()
        return {u["id"]: u for u in result["users"]}

    def get_matches(self, match: str) -> list[dict]:
        """Get user data based on provided info."""
        fields = ["real_name", "nick", "email", "name", "id"]
        response = self._get(
            "user", {"match": match, "include_fields": ",".join(fields)}
        )
        if response.status_code != 200:
            logger.error(f"Error searching {match}: {response.status_code}")
            return

        result = response.json()
        return result["users"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Utility helper functions."""

from __future__ import annotations

from packaging.version import Version
from pathlib import Path
from xml.etree import ElementTree
import logging
import re

import requests

from mots import __version__

logger = logging.getLogger(__name__)

REAL_NAME_RE = re.compile(r"^(?P<name>[\w\ \-]*)?\ ?(?P<info>\W.*)?$")


def generate_machine_readable_name(display_name, keep_case=False):
    """Turn spaces into underscores, and lower the case. Strip all but alphanumerics."""
    words = [w.strip() for w in display_name.split(" ")]
    if keep_case:
        alnum_words = ["".join([c for c in word if c.isalnum()]) for word in words]
    else:
        alnum_words = [
            "".join([c.lower() for c in word if c.isalnum()]) for word in words
        ]
    return "_".join([w for w in alnum_words if w])


def get_list_input(text: str):
    """Parse comma separated list in user input into a list.

    :param test: the text to prompt the user with
    """
    user_input = input(f"{text}: ").split(",")
    return [e.strip() for e in user_input if e]


def parse_real_name(real_name):
    """Parse real_name into name and info."""
    match = REAL_NAME_RE.match(real_name)
    if not match:
        return {"name": "", "info": real_name}
    data = match.groupdict()
    return {k: v.strip() if v else "" for k, v in data.items()}


def mkdir_if_not_exists(path: Path):
    """Check if a directory exists, if not, create it."""
    if not path.exists():
        path.mkdir()
    elif path.exists() and not path.is_dir():
        logger.warning(f"{path} exists but is not a directory.")


def touch_if_not_exists(path: Path):
    """Check if a file exists, if not, create it."""
    if not path.exists():
        path.touch()
    elif path.exists() and not path.is_file():
        logger.warning(f"{path} exists but is not a file.")


def check_for_updates(include_pre_releases: bool = False) -> Version | None:
    """
    Show a message if there is a newer version available.

    This method checks the RSS feed on PyPI.
    """
    URL = "https://pypi.org/rss/project/mots/releases.xml"
    response = requests.get(URL)
    result = ElementTree.fromstring(response.content)
    items = result[0].findall("item")
    versions = [Version(item[0].text) for item in items]

    if not include_pre_releases:
        versions = [v for v in versions if not v.is_prerelease]

    newest_version = max(versions)
    if Version(__version__) < newest_version:
        logger.warning(f"A new version of mots is available ({newest_version})")
        logger.warning(f"You are running {__version__}")
        return newest_version
    else:
        logger.debug(f"You are running the latest version of mots ({__version__})")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Test various CLI commands."""

from mots.cli import version
import subprocess


def test__cli__version():
    out = subprocess.check_output(["mots", "--version"])
    assert out.decode("utf-8").strip() == version()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Module that provides helpers to interact with the PMO API."""

import requests
import logging

from mots.settings import settings


logger = logging.getLogger(__name__)


class PMOClient:
    """A thin wrapper as a PMO API client."""

    def __init__(self, token: str = None, base_url: str = settings.PMO_URL):
        if not token:
            token = settings.PMO_COOKIE
            if not token:
                raise ValueError(
                    "PMO cookies is missing and no other "
                    "token was explicitly provided"
                )
        self.headers = {"Cookie": token, "User-Agent": settings.USER_AGENT}
        self.base_url = base_url

    def _get(self, path: str, params=None):
        params = params or {}
        response = requests.get(
            f"{self.base_url}/{path}", headers=self.headers, params=params
        )
        return response

    def search(self, email: str):
        """Search PMO for a particular email."""
        params = {"q": email, "w": "all"}
        path = "search/simple/"

        return self._get(path, params).json()
